"""Tests for `greenguard.benchmark` module."""
import numpy as np

from greenguard.benchmark import evaluate_templates
from greenguard.demo import load_demo


def test_predict():
    # setup
    templates = [
        'probability.unstack_lstm_timeseries_classifier'
    ]

    window_size_rule = [
        ('1d', '1h')
    ]

    target_times, readings = load_demo()
    target_times = target_times.head(40)
    readings = readings.head(100)

    # run
    scores_df = evaluate_templates(
        target_times=target_times,
        readings=readings,
        templates=templates,
        window_size_rule=window_size_rule,
        tuning_iterations=1,
        cv_splits=2
    )

    # assert
    expected_columns = [
        'problem_name',
        'window_size',
        'resample_rule',
        'template',
        'default_test',
        'default_cv',
        'tuned_cv',
        'tuned_test',
        'tuning_metric',
        'tuning_metric_kwargs',
        'fit_predict_time',
        'default_cv_time',
        'average_cv_time',
        'total_time',
        'status',
        'accuracy_threshold/0.5',
        'f1_threshold/0.5',
        'fpr_threshold/0.5',
        'tpr_threshold/0.5',
    ]

    expected_dtypes = [
        np.dtype('O'),
        np.dtype('O'),
        np.dtype('O'),
        np.dtype('O'),
        np.dtype('float64'),
        np.dtype('float64'),
        np.dtype('float64'),
        np.dtype('float64'),
        np.dtype('O'),
        np.dtype('O'),
        np.dtype('<m8[ns]'),
        np.dtype('<m8[ns]'),
        np.dtype('<m8[ns]'),
        np.dtype('<m8[ns]'),
        np.dtype('O'),
        np.dtype('float64'),
        np.dtype('float64'),
        np.dtype('float64'),
        np.dtype('float64')
    ]

    assert (scores_df.columns.to_list() == expected_columns)
    assert (scores_df.tuned_test.notnull)
    assert (scores_df.dtypes.to_list() == expected_dtypes)



"""Targets module.

This module contains functions to work with target_times.
"""

import logging
import warnings

import numpy as np
import pandas as pd
from tqdm.auto import trange

LOGGER = logging.getLogger(__name__)


def make_targets(target_times, window_size, target, new_targets=None):
    target_times = target_times.sort_values('cutoff_time', ascending=True)
    cutoff_times = target_times.cutoff_time
    window_size = pd.to_timedelta(window_size)
    original_size = len(target_times)
    current_size = original_size
    new_targets = new_targets or current_size

    for index in trange(len(cutoff_times) - 1):
        timestamp = cutoff_times.iloc[index]
        next_time = cutoff_times.iloc[index + 1]

        if timestamp + (window_size * 2) >= next_time:
            continue

        span_start = timestamp + window_size
        span_end = next_time - window_size
        span_length = (span_end - span_start).total_seconds()

        delay = pd.to_timedelta(np.random.randint(span_length), unit='s')
        cutoff_time = span_start + delay

        target_times = target_times.append(pd.Series({
            'turbine_id': target_times.iloc[index].turbine_id,
            'cutoff_time': cutoff_time,
            'target': target
        }), ignore_index=True)

        current_size = len(target_times)
        if current_size == original_size + new_targets:
            return target_times.sort_values('cutoff_time', ascending=True)

    if current_size == original_size:
        warnings.warn('There is no space left between to add more targets.')
        return target_times

    new_targets = new_targets - (current_size - original_size)
    return make_targets(target_times, window_size, target, new_targets)


def _to_timedelta(specification):
    if isinstance(specification, int):
        specification = '{}s'.format(specification)

    return pd.to_timedelta(specification)


def make_target_times(failure_dates, step, start=None, end=None, forecast_window=0,
                      prediction_window=0, before=0, after=0, offset=0, max_true=None,
                      max_false=None, shuffle=True):

    step = _to_timedelta(step)
    start = start or failure_dates.timestamp.min()
    start = start or failure_dates.min()

    forecast_window = _to_timedelta(forecast_window)
    prediction_window = _to_timedelta(prediction_window)
    before = _to_timedelta(before)
    after = _to_timedelta(after)
    offset = _to_timedelta(offset)

    target_times = pd.DataFrame()
    turbines = failure_dates.turbine_id.unique()
    failures = failure_dates.set_index(['turbine_id', 'date'])

    for turbine in turbines:
        turbine_failures = failures.loc[turbine]

        min_failure_date = turbine_failures.index.min() - before
        last_failure_date = turbine_failures.index.max() + after
        turbine_targets = list()
        while min_failure_date < last_failure_date:
            max_failure_date = min_failure_date + prediction_window
            day_failures = turbine_failures.loc[min_failure_date:max_failure_date]

            min_failure_date = min_failure_date + offset

            turbine_targets.append({
                'turbine_id': turbine,
                'target': int(bool(len(day_failures))),
                'cutoff_time': min_failure_date - forecast_window
            })

        turbine_targets = pd.DataFrame(turbine_targets)
        failed = turbine_targets[turbine_targets.target == 1]
        target_times = target_times.append(failed)

        non_failed = turbine_targets[turbine_targets.target == 0]
        non_failed = non_failed.sample(min(max_false, len(non_failed)))

        target_times = target_times.append(non_failed)

    if shuffle:
        target_times = target_times.sample(len(target_times))

    return target_times


def _valid_targets(timestamps):
    def apply_function(row):
        cutoff = row.cutoff_time
        try:
            times = timestamps.loc[row.turbine_id]
        except KeyError:
            return False

        return times['min'] <= cutoff <= times['max']

    return apply_function


def select_valid_targets(target_times, readings, window_size, rule=None):
    """Filter out target_times without enough data for this window_size.

    The table_times table is scanned and checked against the readings table
    considering the window_size. All the target times entries that do not
    have enough data are dropped.

    Args:
        target_times (pandas.DataFrame):
            Target times table, with at least turbined_id and cutoff_time fields.
        readings (pandas.DataFrame):
            Readings table, with at least turbine_id, signal_id, and timestamp ields.
        window_size (str or pandas.TimeDelta):
            TimeDelta specification that indicates the lenght of the training window.
        rule (str or pandas.TimeDelta):
            Resampling rule specification. If given, add that to the max timestamp
            to ensure the period is completely covered.

    Returns:
        pandas.DataFrame:
            New target_times table without the invalid targets.
    """

    timestamps = readings.groupby('turbine_id').timestamp.agg(['min', 'max'])
    timestamps['min'] += pd.to_timedelta(window_size)

    if rule is not None:
        timestamps['max'] += pd.to_timedelta(rule)

    valid = target_times.apply(_valid_targets(timestamps), axis=1)
    valid_targets = target_times[valid]

    length = len(valid_targets)
    LOGGER.info('Dropped %s targets without enough data. Final target_times size: %s',
                len(target_times) - length, length)

    return valid_targets


def drop_duplicates(target_times):
    length = len(target_times)
    filtered = target_times.drop_duplicates()
    new_length = len(filtered)
    if length != new_length:
        LOGGER.warn('Dropped %s duplicate targets!', length - new_length)

    filtered = filtered.drop_duplicates(subset=['turbine_id', 'cutoff_time'], keep=False)
    final_length = len(filtered)
    if new_length != final_length:
        LOGGER.warn('Dropped %s incoherent targets!', new_length - final_length)

    return filtered.copy()



# -*- coding: utf-8 -*-

import logging

from mlblocks import MLPipeline


def clone_pipeline(pipeline):
    return MLPipeline.from_dict(pipeline.to_dict())


def walk(document, transform):
    if not isinstance(document, dict):
        return document

    new_doc = dict()
    for key, value in document.items():
        if isinstance(value, dict):
            value = walk(value, transform)
        elif isinstance(value, list):
            value = [walk(v, transform) for v in value]

        new_key, new_value = transform(key, value)
        new_doc[new_key] = new_value

    return new_doc


def remove_dots(document):
    return walk(document, lambda key, value: (key.replace('.', '-'), value))


def restore_dots(document):
    return walk(document, lambda key, value: (key.replace('-', '.'), value))


def logging_setup(verbosity=1, logfile=None, logger_name=None):
    logger = logging.getLogger(logger_name)
    log_level = (3 - verbosity) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(module)s - %(message)s'
    formatter = logging.Formatter(fmt)
    logger.setLevel(log_level)
    logger.propagate = False

    if logfile:
        file_handler = logging.FileHandler(logfile)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    else:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)


def as_list(param):
    """Make sure that param is a ``list``."""
    if isinstance(param, (list, tuple)):
        return param

    return [param]



import logging
import os

import dask
import pandas as pd

from greenguard.targets import drop_duplicates, select_valid_targets

LOGGER = logging.getLogger(__name__)


class CSVLoader:
    """Load the required readings from CSV files.

    The CSVLoader class is responsible for analyzing the target_times table
    and then load the required readings from CSV files.

    Also, optionally, it can perform a resampling aggregation while loading
    the data, reducing the amount of memory requirements.

    The CSVLoader class uses Dask to parallelize all the IO and resampling
    computation and reduce loading times.

    Args:
        readings_path (str):
            Path to the readings folder, where a folder exist for each turbine.
        rule (str):
            Resampling rule, as expected by ``DataFrame.resmple``. The rule is a
            string representation of a TimeDelta, which includes a number and a
            unit. For example: ``3d``, ``1w``, ``6h``.
            If ``None``, resampling is disabled.
        aggregation (str):
            Name of the aggregation to perform during the resampling.
        unstack (bool):
            Whether to unstack the resampled data, generating one column per signal.
            Only used when resampling. Defaults to ``False``.
    """

    DEFAULT_DATETIME_FMT = '%m/%d/%y %H:%M:%S'
    DEFAULT_FILENAME_FMT = '%Y-%m.csv'

    def __init__(self, readings_path='.', rule=None, aggregation='mean', unstack=False,
                 datetime_fmt=DEFAULT_DATETIME_FMT, filename_fmt=DEFAULT_FILENAME_FMT):
        self._readings_path = readings_path
        self._rule = rule
        self._aggregation = aggregation
        self._unstack = unstack
        self._datetime_fmt = datetime_fmt
        self._filename_fmt = filename_fmt

    @dask.delayed
    def __filter_by_signal(self, readings, signals):
        if signals is not None:
            LOGGER.debug('Filtering by signal')
            readings = readings[readings.signal_id.isin(signals)]

        try:
            readings['value'] = readings['value'].astype(float)
        except ValueError:
            signals = readings[readings['value'].str.isnumeric()].signal_id.unique()
            raise ValueError('Signals contain non-numerical values: {}'.format(signals))

        LOGGER.debug('Selected %s readings by signal', len(readings))

        return readings.copy()

    @dask.delayed
    def __filter_by_timestamp(self, readings, timestamps):
        LOGGER.debug('Parsing timestamps')
        readings_ts = pd.to_datetime(readings['timestamp'], format=self._datetime_fmt)
        readings['timestamp'] = readings_ts

        LOGGER.debug('Filtering by timestamp')

        related = [False] * len(readings)
        for row in timestamps.itertuples():
            lower = row.start <= readings_ts
            upper = readings_ts <= row.stop
            related |= lower & upper

        readings = readings[related]

        LOGGER.debug('Selected %s readings by timestamp', len(readings))

        return readings.copy()

    @dask.delayed
    def __load_readings_file(self, turbine_file, timestamps, signals):
        LOGGER.debug('Loading file %s', turbine_file)
        data = pd.read_csv(turbine_file, low_memory=False)
        data.columns = data.columns.str.lower()
        data = data.rename(columns={'signal': 'signal_id'})

        if 'unnamed: 0' in data.columns:
            # Someone forgot to drop the index before
            # storing the DataFrame as a CSV
            del data['unnamed: 0']

        LOGGER.debug('Loaded %s readings from file %s', len(data), turbine_file)

        return data

    @dask.delayed
    def __consolidate(self, readings, turbine_id):
        readings = pd.concat(readings, ignore_index=True)
        readings.insert(0, 'turbine_id', turbine_id)

        LOGGER.info('Loaded %s readings from turbine %s', len(readings), turbine_id)

        return readings

    def _get_filenames(self, turbine_path, timestamps):
        min_csv = timestamps.start.dt.strftime(self._filename_fmt)
        max_csv = timestamps.stop.dt.strftime(self._filename_fmt)

        for filename in sorted(os.listdir(turbine_path)):
            if ((min_csv <= filename) & (filename <= max_csv)).any():
                yield os.path.join(turbine_path, filename)

    @staticmethod
    def _join_names(names):
        """Join the names of a multi-level index with an underscore."""

        levels = (str(name) for name in names if name != '')
        return '_'.join(levels)

    @dask.delayed
    def __resample(self, readings):
        LOGGER.info('Resampling: %s - %s', self._rule, self._aggregation)
        grouped = readings.groupby('signal_id')
        dfr = grouped.resample(rule=self._rule, on='timestamp')
        agg = dfr.agg(self._aggregation)

        LOGGER.info('%s readings reduced to %s', len(readings), len(agg))

        if self._unstack:
            agg = agg.unstack(level='signal_id').reset_index()
            agg.columns = agg.columns.map(self._join_names)
            return agg
        else:
            return agg.reset_index()

    def _load_turbine(self, turbine_id, timestamps, signals=None):
        if 'turbine_id' in timestamps:
            timestamps = timestamps[timestamps.turbine_id == turbine_id]

        turbine_path = os.path.join(self._readings_path, turbine_id)
        filenames = self._get_filenames(turbine_path, timestamps)

        readings = list()
        for filename in filenames:
            file_readings = self.__load_readings_file(filename, timestamps, signals)
            file_readings = self.__filter_by_signal(file_readings, signals)
            file_readings = self.__filter_by_timestamp(file_readings, timestamps)

            if self._rule:
                file_readings = self.__resample(file_readings)

            readings.append(file_readings)

        if readings:
            readings = self.__consolidate(readings, turbine_id)

        return readings

    @staticmethod
    def _get_timestamps(target_times, window_size):
        cutoff_times = target_times.cutoff_time
        min_times = cutoff_times - window_size

        return pd.DataFrame({
            'turbine_id': target_times.turbine_id,
            'start': min_times,
            'stop': cutoff_times,
        })

    def load(self, target_times, window_size, signals=None, debug=False, select_valid=True):
        """Load the readings needed for the given target_times and window_size.

        Optionally filter the signals that are loaded and discard the rest.

        Args:
            target_times (str or pandas.DataFrame):
                target_times ``DataFrame`` or path to the corresponding CSV file.
                The table must have three volumns, ``turbine_id``, ``target`` and
                ``cutoff_time``.
            window_size (str):
                Amount of data to load before each cutoff time, specified as a string
                representation of a TimeDelta, which includes a number and a
                unit. For example: ``3d``, ``1w``, ``6h``.
            signals (list or pandas.DataFrame):
                List of signal names or table that has a ``signal_id`` column to
                use as the signal names list.
            debug (bool):
                Force single thread execution for easy debugging. Defaults to ``False``.

        Returns:
            pandas.DataFrame:
                Table of readings for the target times, including the columns ``turbine_id``,
                ``signal_id``, ``timestamp`` and ``value``.
        """
        if isinstance(target_times, str):
            target_times = pd.read_csv(target_times)
            target_times['cutoff_time'] = pd.to_datetime(target_times['cutoff_time'])

        target_times = drop_duplicates(target_times)

        if isinstance(signals, pd.DataFrame):
            signals = signals.signal_id

        window_size = pd.to_timedelta(window_size)
        timestamps = self._get_timestamps(target_times, window_size)

        readings = list()
        for turbine_id in timestamps.turbine_id.unique():
            readings.append(self._load_turbine(turbine_id, timestamps, signals))

        dask_scheduler = 'single-threaded' if debug else None
        computed = dask.compute(*readings, scheduler=dask_scheduler)

        found_readings = [c for c in computed if len(c)]
        if not found_readings:
            msg = 'No readings found for the given target times in {}'.format(self._readings_path)
            raise ValueError(msg)

        readings = pd.concat(found_readings, ignore_index=True, sort=False)

        LOGGER.info('Loaded %s turbine readings', len(readings))

        if select_valid:
            target_times = select_valid_targets(target_times, readings, window_size, self._rule)
            return target_times, readings

        return readings



import argparse
import logging
import multiprocessing as mp
import os
import pickle
import re
import sys
import warnings
from datetime import datetime
from itertools import product

import pandas as pd
import tabulate
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from greenguard import get_pipelines
from greenguard.demo import load_demo
from greenguard.loaders import CSVLoader
from greenguard.metrics import (METRICS, accuracy_score, f1_score,
                                fpr_score, tpr_score, threshold_score)
from greenguard.pipeline import GreenGuardPipeline, generate_init_params, generate_preprocessing
from greenguard.results import load_results, write_results

LOGGER = logging.getLogger(__name__)

DEFAULT_TUNING_METRIC_KWARGS = {'threshold': 0.5}
LEADERBOARD_COLUMNS = [
    'problem_name',
    'window_size',
    'resample_rule',
    'template',
    'default_test',
    'default_cv',
    'tuned_cv',
    'tuned_test',
    'tuning_metric',
    'tuning_metric_kwargs',
    'fit_predict_time',
    'default_cv_time',
    'average_cv_time',
    'total_time',
    'status',
]


def _scorer(metric, metric_args):
    if isinstance(metric, str):
        metric, cost = METRICS[metric]

    def f(expected, observed):
        try:
            return metric(expected, observed, **metric_args)
        except TypeError:
            if 'threshold' not in metric_args:
                raise

            kwargs = metric_args.copy()
            threshold = kwargs.pop('threshold')
            observed = observed >= threshold
            return metric(expected, observed, **kwargs)

    return f


def _build_init_params(template, window_size, rule, template_params):
    if 'dfs' in template:
        window_size_rule_params = {
            'pandas.DataFrame.resample#1': {
                'rule': rule,
            },
            'featuretools.dfs.json#1': {
                'training_window': window_size,
            }
        }
    elif 'lstm' in template:
        window_size_rule_params = {
            'pandas.DataFrame.resample#1': {
                'rule': rule,
            },
            'mlprimitives.custom.timeseries_preprocessing.cutoff_window_sequences#1': {
                'window_size': window_size,
            }
        }

    for primitive, params in window_size_rule_params.items():
        primitive_params = template_params.setdefault(primitive, {})
        primitive_params.update(params)

    return template_params


def evaluate_template(
    template,
    target_times,
    readings,
    tuning_iterations=50,
    init_params=None,
    preprocessing=0,
    metrics=None,
    threshold=None,
    tpr=None,
    tuning_metric='roc_auc_score',
    tuning_metric_kwargs=DEFAULT_TUNING_METRIC_KWARGS,
    cost=False,
    cv_splits=3,
    test_size=0.25,
    random_state=0,
    cache_path=None,
    scores={}
):
    """Returns the scores for a given template.

    Args:
        template (str):
            Given template to evaluate.
        target_times (DataFrame):
            Contains the specefication problem that we are solving, which has three columns:

                * turbine_id: Unique identifier of the turbine which this label corresponds to.
                * cutoff_time: Time associated with this target.
                * target: The value that we want to predict. This can either be a numerical value
                          or a categorical label. This column can also be skipped when preparing
                          data that will be used only to make predictions and not to fit any
                          pipeline.

        readings (DataFrame):
            Contains the signal data from different sensors, with the following columns:

                * turbine_id: Unique identifier of the turbine which this reading comes from.
                * signal_id: Unique identifier of the signal which this reading comes from.
                * timestamp (datetime): Time where the reading took place, as a datetime.
                * value (float): Numeric value of this reading.

        metric (function or str):
            Metric to use. If an ``str`` is give it must be one of the metrics
            defined in the ``greenguard.metrics.METRICS`` dictionary.
        tuning_iterations (int):
            Number of iterations to be used.
        preprocessing (int, list or dict):
            Number of preprocessing steps to be used.
        init_params (list):
            Initialization parameters for the pipeline.
        cost (bool):
            Wheter the metric is a cost function (the lower the better) or not.
        test_size (float):
            Percentage of the data set to be used for the test.
        cv_splits (int):
            Amount of splits to create.
        random_state (int):
            Random number of train_test split.
        cache_path (str):
            If given, cache the generated cross validation splits in this folder.
            Defatuls to ``None``.

    Returns:
        scores (dict):
            Stores the four types of scores that are being evaluate.
    """
    start_time = datetime.utcnow()
    scores['tuning_metric'] = str(tuning_metric)
    scores['tuning_metric_kwargs'] = tuning_metric_kwargs
    tuning_metric = _scorer(tuning_metric, tuning_metric_kwargs)

    train, test = train_test_split(target_times, test_size=test_size, random_state=random_state)

    pipeline = GreenGuardPipeline(
        template,
        metric=tuning_metric,
        cost=cost,
        cv_splits=cv_splits,
        init_params=init_params,
        preprocessing=preprocessing,
        cache_path=cache_path
    )

    # Computing the default test score
    fit_predict_time = datetime.utcnow()
    pipeline.fit(train, readings)
    predictions = pipeline.predict(test, readings)
    fit_predict_time = datetime.utcnow() - fit_predict_time

    scores['default_test'] = tuning_metric(test['target'], predictions)

    # Computing the default cross validation score
    default_cv_time = datetime.utcnow()
    session = pipeline.tune(train, readings)
    session.run(1)
    default_cv_time = datetime.utcnow() - default_cv_time

    scores['default_cv'] = pipeline.cv_score

    # Computing the cross validation score with tuned hyperparameters
    average_cv_time = datetime.utcnow()
    session.run(tuning_iterations)
    average_cv_time = (datetime.utcnow() - average_cv_time) / tuning_iterations

    scores['tuned_cv'] = pipeline.cv_score

    # Computing the test score with tuned hyperparameters
    pipeline.fit(train, readings)
    predictions = pipeline.predict(test, readings)
    ground_truth = test['target']

    # compute different metrics
    if tpr:
        tpr = tpr if isinstance(tpr, list) else [tpr]
        for value in tpr:
            threshold = threshold_score(ground_truth, predictions, tpr)
            scores[f'fpr_tpr/{value}'] = fpr_score(ground_truth, predictions, tpr=tpr)
            predictions_classes = predictions >= threshold
            scores[f'accuracy_tpr/{value}'] = accuracy_score(ground_truth, predictions_classes)
            scores[f'f1_tpr/{value}'] = f1_score(ground_truth, predictions_classes)
            scores[f'threshold_tpr/{value}'] = threshold_score(ground_truth, predictions, value)

            if f'accuracy_tpr/{value}' not in LEADERBOARD_COLUMNS:
                LEADERBOARD_COLUMNS.extend([
                    f'accuracy_tpr/{value}',
                    f'f1_tpr/{value}',
                    f'fpr_tpr/{value}',
                    f'threshold_tpr/{value}',
                ])

    else:
        threshold = 0.5 if threshold is None else threshold
        threshold = threshold if isinstance(threshold, list) else [threshold]

        for value in threshold:
            scores[f'fpr_threshold/{value}'] = fpr_score(
                ground_truth, predictions, threshold=value)

            predictions_classes = predictions >= threshold
            scores[f'accuracy_threshold/{value}'] = accuracy_score(
                ground_truth, predictions_classes)

            scores[f'f1_threshold/{value}'] = f1_score(ground_truth, predictions_classes)
            scores[f'tpr_threshold/{value}'] = tpr_score(ground_truth, predictions, value)

            if f'accuracy_threshold/{value}' not in LEADERBOARD_COLUMNS:
                LEADERBOARD_COLUMNS.extend([
                    f'accuracy_threshold/{value}',
                    f'f1_threshold/{value}',
                    f'fpr_threshold/{value}',
                    f'tpr_threshold/{value}',
                ])

    scores['tuned_test'] = tuning_metric(test['target'], predictions)
    scores['fit_predict_time'] = fit_predict_time
    scores['default_cv_time'] = default_cv_time
    scores['average_cv_time'] = average_cv_time
    scores['total_time'] = datetime.utcnow() - start_time

    return scores


def evaluate_templates(
    templates,
    window_size_rule,
    tuning_iterations=50,
    init_params=None,
    preprocessing=0,
    metrics=None,
    threshold=None,
    tpr=None,
    tuning_metric='roc_auc_score',
    tuning_metric_kwargs=DEFAULT_TUNING_METRIC_KWARGS,
    target_times=None,
    readings=None,
    cost=False,
    test_size=0.25,
    cv_splits=3,
    random_state=0,
    cache_path=None,
    cache_results=None,
    problem_name=None,
    output_path=None,
    progress_bar=None,
    multiprocess=False
):
    """Execute the benchmark process and optionally store the result as a ``CSV``.

    Args:
        templates (list):
            List of templates to try.
        window_size_rule (list):
            List of tuples (int, str or Timedelta object).
        metric (function or str):
            Metric to use. If an ``str`` is give it must be one of the metrics
            defined in the ``greenguard.metrics.METRICS`` dictionary.
        tuning_iterations (int):
            Number of iterations to be used.
        init_params (dict):
            Initialization parameters for the pipelines.
        target_times (DataFrame):
            Contains the specefication problem that we are solving, which has three columns:

                * turbine_id: Unique identifier of the turbine which this label corresponds to.
                * cutoff_time: Time associated with this target.
                * target: The value that we want to predict. This can either be a numerical value
                          or a categorical label. This column can also be skipped when preparing
                          data that will be used only to make predictions and not to fit any
                          pipeline.

        readings (DataFrame):
            Contains the signal data from different sensors, with the following columns:

                * turbine_id: Unique identifier of the turbine which this reading comes from.
                * signal_id: Unique identifier of the signal which this reading comes from.
                * timestamp (datetime): Time where the reading took place, as a datetime.
                * value (float): Numeric value of this reading.

        preprocessing (int, list or dict):
            Number of preprocessing steps to be used.
        cost (bool):
            Wheter the metric is a cost function (the lower the better) or not.
        test_size (float):
            Percentage of the data set to be used for the test.
        cv_splits (int):
            Amount of splits to create.
        random_state (int):
            Random number of train_test split.
        output_path (str):
            Path where to save the benchmark report.
        cache_path (str):
            If given, cache the generated cross validation splits in this folder.
            Defatuls to ``None``.

    Returns:
        pandas.DataFrame or None:
            If ``output_path`` is ``None`` it will return a ``pandas.DataFrame`` object,
            else it will dump the results in the specified ``output_path``.

    Example:
        >>> from sklearn.metrics import f1_score
        >>> templates = [
        ...    'normalize_dfs_xgb_classifier',
        ...    'unstack_lstm_timeseries_classifier'
        ... ]
        >>> window_size_rule = [
        ...     ('30d','12h'),
        ...     ('7d','4h')
        ... ]
        >>> preprocessing = [0, 1]
        >>> scores_df = evaluate_templates(
        ...                 templates=templates,
        ...                 window_size_rule=window_size_rule,
        ...                 metric=f1_score,
        ...                 tuning_iterations=5,
        ...                 preprocessing=preprocessing,
        ...                 cost=False,
        ...                 test_size=0.25,
        ...                 cv_splits=3,
        ...                 random_state=0
        ...             )
        >>> scores_df
                                 template window_size resample_rule  default_test  default_cv  tuned_cv  tuned_test status
    0  unstack_lstm_timeseries_classifier         30d           12h      0.720000    0.593634  0.627883    0.775510     OK
    1  unstack_lstm_timeseries_classifier          7d            4h      0.723404    0.597440  0.610766    0.745098     OK
    2        normalize_dfs_xgb_classifier         30d           12h      0.581818    0.619698  0.637123    0.596491     OK
    3        normalize_dfs_xgb_classifier          7d            4h      0.581818    0.619698  0.650367    0.603774     OK

    """  # noqa
    if readings is None and target_times is None:
        target_times, readings = load_demo()

    init_params = generate_init_params(templates, init_params)
    preprocessing = generate_preprocessing(templates, preprocessing)

    scores_list = []
    for template, window_rule in product(templates, window_size_rule):
        window_size, rule = window_rule


        try:
            LOGGER.info('Evaluating template %s on problem %s (%s, %s)',
                        template, problem_name, window_size, rule)

            template_params = init_params[template]
            template_params = _build_init_params(template, window_size, rule, template_params)
            template_preprocessing = preprocessing[template]
            if multiprocess:
                manager = mp.Manager()
                scores = manager.dict()
                process = mp.Process(
                    target=evaluate_template,
                    args=(
                        template,
                        target_times,
                        readings,
                        tuning_iterations,
                        init_params,
                        preprocessing,
                        metrics,
                        threshold,
                        tpr,
                        tuning_metric,
                        tuning_metric_kwargs,
                        cost,
                        cv_splits,
                        test_size,
                        random_state,
                        cache_path,
                        scores
                    )
                )

                process.start()
                process.join()
                if 'tuned_test' not in scores:
                    scores['status'] = 'ERRORED'

                scores = dict(scores)  # parse the managed dict to dict for pandas.

            else:
                scores = dict()
                scores['problem_name'] = problem_name
                scores['template'] = template
                scores['window_size'] = window_size
                scores['resample_rule'] = rule
                result = evaluate_template(
                    template=template,
                    target_times=target_times,
                    readings=readings,
                    metrics=metrics,
                    tuning_metric=tuning_metric,
                    tuning_metric_kwargs=tuning_metric_kwargs,
                    threshold=threshold,
                    tpr=tpr,
                    tuning_iterations=tuning_iterations,
                    preprocessing=template_preprocessing,
                    init_params=template_params,
                    cost=cost,
                    test_size=test_size,
                    cv_splits=cv_splits,
                    random_state=random_state,
                    cache_path=cache_path
                )

                scores.update(result)
                scores['status'] = 'OK'

        except Exception:
            scores['status'] = 'ERRORED'
            LOGGER.exception('Could not score template %s ', template)

        if cache_results:
            os.makedirs(cache_results, exist_ok=True)
            template_name = template
            if os.path.isfile(template_name):
                template_name = os.path.basename(template_name).replace('.json', '')

            file_name = '{}_{}_{}_{}.csv'.format(problem_name, template_name, window_size, rule)
            df = pd.DataFrame([scores]).reindex(LEADERBOARD_COLUMNS, axis=1)
            df.to_csv(os.path.join(cache_results, file_name), index=False)

        scores_list.append(scores)

        if progress_bar:
            progress_bar.update(1)

    results = pd.DataFrame.from_records(scores_list)
    results = results.reindex(LEADERBOARD_COLUMNS, axis=1)

    if output_path:
        LOGGER.info('Saving benchmark report to %s', output_path)
        results.to_csv(output_path)
    else:
        return results


def _generate_target_times_readings(target_times, readings_path, window_size, rule, signals):
    """
    Returns:
        pandas.DataFrame:
            Table of readings for the target times, including the columns ``turbine_id``,
            ``signal_id``, ``timestamp`` and ``value``.
    """
    csv_loader = CSVLoader(
        readings_path,
        rule=rule,
    )

    return csv_loader.load(target_times, window_size=window_size, signals=signals)


def make_problems(target_times_paths, readings_path, window_size_resample_rule,
                  output_path=None, signals=None):
    """Make problems with the target times and readings for each window size and resample rule.

    Create problems in the accepted format by ``run_benchmark`` as pickle files containing:

        * ``target_times``: ``pandas.DataFrame`` containing the target times.
        * ``readings``: ``pandas.DataFrame`` containing the readings for the target times.
        * ``window_size``: window size value used.
        * ``resample_rule``: resample rule value used.

    Or return a ``dict`` containing as keys the names of the problems generated and tuples with
    the previously specified fields of target times, readings, window size and resample rule.

    Args:
        target_times_paths (list):
            List of paths to CSVs that contain target times.
        readings_path (str):
            Path to the folder where readings in raw CSV format can be found.
        window_size_resample_rule (list):
            List of tuples (int, str or Timedelta object).
        output_path (str):
            Path to save the generated problems.
        signals (str):
            List of signal names or csv file that has a `signal_id` column to use as the signal
            names list.
    """
    if isinstance(target_times_paths, str):
        target_times_paths = [target_times_paths]
    if isinstance(target_times_paths, list):
        target_times_paths = {
            os.path.basename(path).replace('.csv', ''): path
            for path in target_times_paths
        }

    if output_path:
        generated_problems = list()
    else:
        generated_problems = {}

    if isinstance(signals, str) and os.path.exists(signals):
        signals = pd.read_csv(signals).signal_id

    for problem_name, target_time_path in tqdm(target_times_paths.items()):
        for window_size, rule in window_size_resample_rule:
            target_times = pd.read_csv(target_time_path, parse_dates=['cutoff_time'])
            new_target_times, readings = _generate_target_times_readings(
                target_times,
                readings_path,
                window_size,
                rule,
                signals=signals,
            )

            pickle_name = '{}_{}_{}'.format(problem_name, window_size, rule)

            if output_path:
                os.makedirs(output_path, exist_ok=True)
                output_pickle_path = os.path.join(output_path, pickle_name + '.pkl')
                with open(output_pickle_path, 'wb') as pickle_file:
                    pickle.dump((new_target_times, readings, window_size, rule), pickle_file)

                generated_problems.append(output_pickle_path)

            else:
                generated_problems[pickle_name] = (new_target_times, readings, window_size, rule)

    return generated_problems


def run_benchmark(templates, problems, window_size_resample_rule=None,
                  tuning_iterations=50, signals=None, preprocessing=0, init_params=None,
                  metrics=None, threshold=None, tpr=None, tuning_metric='roc_auc_score',
                  tuning_metric_kwargs=DEFAULT_TUNING_METRIC_KWARGS, cost=False, cv_splits=5,
                  test_size=0.33, random_state=0, cache_path=None, cache_results=None,
                  output_path=None, multiprocess=False):
    """Execute the benchmark function and optionally store the result as a ``CSV``.

    This function provides a user-friendly interface to interact with the ``evaluate_templates``
    function. It allows the user to specify an ``output_path`` where the results can be
    stored. If this path is not provided, a ``pandas.DataFrame`` will be returned.

    This function evaluates each template against each problem for each window size and resample
    rule possible, and will tune each teamplate for the given amount of tuning iterations.

    The problems can be a pickle file that contains the following values:

        * ``target_times``: ``pandas.DataFrame`` containing the target times.
        * ``readings``: ``pandas.DataFrame`` containing the readings for the target times.
        * ``window_size``: window size value used.
        * ``resample_rule``: resample rule value used.

    Or it can be dictionary containing the problem's name and as values either a path to a pickle
    file or a tuple containing the previously specified fields.

    Args:
        templates (str or list):
            Name of the json pipelines that will be evaluated against the problems.
        problems (str, list or dict):
            There are three possible values for problems:

                * ``str``: Path to a given problem stored as a pickle file (pkl).
                * ``list``: List of paths to given problems stored as a pickle files (pkl).
                * ``dict``: A dict containing as keys the name of the problem and as value the
                            path to a pickle file or a tuple with target times and readings data
                            frames and the window size and resample rule used to generate this
                            problem.

            The pickle files has to contain a tuple with target times and readings data frames and
            the window size and resample rule used to generate that problem. We recommend using
            the function ``make_problems`` to generate those files.

        window_size_resample_rule (list):
            List of tuples (int, str or Timedelta object).
        tuning_iterations (int):
            Amount of tuning iterations to perfrom over each template.
        signals (str or list):
            Path to a csv file containing ``signal_id`` column that we would like to use or a
            ``list`` of signals that we would like to use. If ``None`` use all the signals from
            the readings.
        preprocessing (int, dict or list):
            There are three possible values for preprocessing:

                * ``int``: the value will be used for all templates.
                * ``dict`` with the template name as a key and a number as a value, will
                  be used for that template.
                * ``list``: each value will be assigned to the corresponding position of
                  self.templates.

            Defaults to ``0``.
        init_params (dict or list):
            There are three possible values for init_params:

                * Init params ``dict``: It will be used for all templates.
                * ``dict`` with the name of the template as a key and dictionary with its
                  init params.
                * ``list``: each value will be assigned to the corresponding position of
                  self.templates.

            Defaults to ``None``.
        metric (function or str):
            Metric to use. If an ``str`` is give it must be one of the metrics
            defined in the ``greenguard.metrics.METRICS`` dictionary.
        cost (bool):
            Whether the metric is a cost function (the lower the better) or not.
            Defaults to ``False``.
        cv_splits (int):
            Number of cross validation folds to use. Defaults to ``5``.
        test_size (float):
            Amount of data that will be saved for test, represented in percentage between 0 and 1.
        random_state (int or RandomState):
            random state to use for the cross validation partitioning. Defaults to ``0``.
        cache_path (str):
            If given, cache the generated cross validation splits in this folder.
            Defatuls to ``None``.
        cache_results (str):
            If provided, store the progress of each pipeline and each problem while runing.
        output_path (str):
            If provided, store the results to the given filename. Defaults to ``None``.
    """
    templates = templates if isinstance(templates, (list, tuple)) else [templates]
    results = list()
    if isinstance(problems, str):
        problems = [problems]
    if isinstance(problems, list):
        problems = {
            os.path.basename(problem).replace('.pkl', ''): problem
            for problem in problems
        }

    if signals is not None:
        if isinstance(signals, str) and os.path.exists(signals):
            signals = pd.read_csv(signals).signal_id

    total_runs = len(templates) * len(problems) * len(window_size_resample_rule or [1])
    pbar = tqdm(total=total_runs)

    for problem_name, problem in problems.items():
        # remove window_size resample_rule nomenclature from the problem's name
        problem_name = re.sub(r'\_\d+[DdHhMmSs]', r'', problem_name)

        if isinstance(problem, str):
            with open(problem, 'rb') as pickle_file:
                target_times, readings, orig_window_size, orig_rule = pickle.load(pickle_file)
        else:
            target_times, readings, orig_window_size, orig_rule = problem

        if signals is not None:
            readings = readings[readings.signal_id.isin(signals)]

        wsrr = window_size_resample_rule or [(orig_window_size, orig_rule)]

        orig_window_size = pd.to_timedelta(orig_window_size)
        orig_rule = pd.to_timedelta(orig_rule)

        for window_size, resample_rule in wsrr:

            # window_size can be only smaller than pickle window size
            # resample rule can be only bigger than picke rule
            if (orig_window_size >= pd.to_timedelta(window_size)
                    and orig_rule <= pd.to_timedelta(resample_rule)): # noqa W503

                df = evaluate_templates(
                    templates,
                    [(window_size, resample_rule)],
                    metrics=metrics,
                    tuning_iterations=tuning_iterations,
                    threshold=threshold,
                    tpr=tpr,
                    init_params=init_params,
                    target_times=target_times,
                    readings=readings,
                    preprocessing=preprocessing,
                    cost=cost,
                    test_size=test_size,
                    cv_splits=cv_splits,
                    random_state=random_state,
                    cache_path=cache_path,
                    cache_results=cache_results,
                    problem_name=problem_name,
                    output_path=None,
                    progress_bar=pbar,
                    multiprocess=multiprocess,
                )

                results.append(df)

                if cache_results:
                    file_name = '{}_{}_{}.csv'.format(problem_name, window_size, resample_rule)
                    df.to_csv(os.path.join(cache_results, file_name), index=False)

            else:
                pbar.update(1)

                msg = 'Invalid window size or resample rule {}.'.format(
                    (window_size, orig_window_size, resample_rule, orig_rule))

                LOGGER.warn(msg)

    pbar.close()

    results = pd.concat(results, ignore_index=True)
    if output_path:
        if os.path.dirname(output_path):
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
        results.to_csv(output_path, index=False)

    else:
        return results


def _setup_logging(args):
    # Logger setup
    log_level = (3 - args.verbose) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(name)s - %(module)s - %(message)s'
    logging.basicConfig(filename=args.logfile, level=log_level, format=fmt)
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("tensorflow").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.CRITICAL)
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # FATAL


def _run(args):
    _setup_logging(args)
    if args.templates is None:
        args.templates = get_pipelines()

    window_size_resample_rule = None
    if args.window_size_resample_rule:
        pattern = re.compile(r'\d+[DdHhMmSs]')
        window_size_resample_rule = [
            tuple(pattern.findall(item))
            for item in args.window_size_resample_rule
        ]

    if args.tuning_metric_kwargs:
        args.tuning_metric_kwargs = json.loads(args.tuning_metric_kwargs)

    else:
        args.tuning_metric_kwargs = DEFAULT_TUNING_METRIC_KWARGS

    # run
    results = run_benchmark(
        templates=args.templates,
        problems=args.problems,
        window_size_resample_rule=window_size_resample_rule,
        cv_splits=args.cv_splits,
        metrics=args.metrics,
        threshold=args.threshold,
        tpr=args.tpr,
        tuning_metric=args.tuning_metric,
        tuning_metric_kwargs=args.tuning_metric_kwargs,
        test_size=args.test_size,
        random_state=args.random_state,
        cache_path=args.cache_path,
        cache_results=args.cache_results,
        tuning_iterations=args.iterations,
        output_path=args.output_path,
        signals=args.signals,
        multiprocess=args.multiprocess
    )

    if not args.output_path:
        print(tabulate.tabulate(
            results,
            tablefmt='github',
            headers=results.columns
        ))


def summarize_results(input_paths, output_path):
    """Load multiple benchmark results CSV files and compile a summary.

    The result is an Excel file with one tab for each results CSV file
    and an additional Number of Wins tab with a summary.

    Args:
        inputs_paths (list[str]):
            List of paths to CSV files where the benchmarks results are stored.
            These files must have one column per Tuner and one row per Challenge.
        output_path (str):
            Path, including the filename, where the Excel file will be created.
    """
    results = load_results(input_paths)
    write_results(results, output_path)


def _summarize_results(args):
    summarize_results(args.input, args.output)


def _make_problems(args):
    window_size_resample_rule = list(product(args.window_size, args.resample_rule))
    make_problems(
        args.target_times_paths,
        args.readings_path,
        window_size_resample_rule,
        output_path=args.output_path,
        signals=args.signals
    )


def _get_parser():
    parser = argparse.ArgumentParser(description='GreenGuard Benchmark Command Line Interface.')
    parser.set_defaults(action=None)
    action = parser.add_subparsers(title='action')
    action.required = True

    # Run action
    run = action.add_parser('run', help='Run the GreenGuard Benchmark')
    run.set_defaults(action=_run)
    run.set_defaults(user=None)

    run.add_argument('-v', '--verbose', action='count', default=0,
                     help='Be verbose. Use -vv for increased verbosity.')
    run.add_argument('-l', '--logfile',
                     help='Log file.')
    run.add_argument('-t', '--templates', nargs='+',
                     help='Perform benchmarking over the given list of templates.')
    run.add_argument('-p', '--problems', nargs='+', required=False,
                     help='Perform benchmarking over a list of pkl problems.')
    run.add_argument('-w', '--window-size-resample-rule', nargs='+', required=False,
                     help='List of window sizes values to benchmark.')
    run.add_argument('-o', '--output_path', type=str,
                     help='Output path where to store the results.')
    run.add_argument('-s', '--cv-splits', type=int, default=5,
                     help='Amount of cross validation splits to use.')
    run.add_argument('-m', '--metrics', nargs='+',
                     help='Names of metric functions to be used for the benchmarking.')
    run.add_argument('-T', '--threshold', nargs='+', type=float,
                     help='Threhshold values for the metrics.')
    run.add_argument('-P', '--tpr', nargs='+', type=float,
                     help='TPR vales for the metrics, if provided threshold will be ignored.')
    run.add_argument('-n', '--random-state', type=int, default=0,
                     help='Random state for the cv splits.')
    run.add_argument('-e', '--test-size', type=float, default=0.33,
                     help='Percentage of the data set to be used for the test.')
    run.add_argument('-c', '--cache-path', type=str,
                     help='Path to cache the generated cross validation splits in.')
    run.add_argument('-R', '--cache-results', type=str,
                     help='Path to store the csv files for each problem and template.')
    run.add_argument('-i', '--iterations', type=int, default=100,
                     help='Number of iterations to perform per challenge with each candidate.')
    run.add_argument('-S', '--signals', type=str,
                     help='Path to csv file that has signal_id column to use as the signal')
    run.add_argument('-k', '--tuning-metric', type=str, default='roc_auc_score',
                     help='Tuning metric to be used.')
    run.add_argument('-K', '--tuning-metric-kwargs', type=str,
                     help='Tuning metric args to be used with the metric.')
    run.add_argument('-u', '--multiprocess', action='store_true',
                     help='Wether or not to spawn a separate process and avoid crashing.')


    # Summarize action
    summary = action.add_parser('summarize-results',
                                help='Summarize the GreenGuard Benchmark results')
    summary.set_defaults(action=_summarize_results)
    summary.add_argument('input', nargs='+', help='Input path with results.')
    summary.add_argument('output', help='Output file.')

    # Make problems action
    problems = action.add_parser('make-problems', help='Create GreenGuard problems')
    problems.set_defaults(action=_make_problems)
    problems.add_argument('target-times-paths', nargs='+', help='List of target times paths.')
    problems.add_argument('readings-path', type=str, help='Path to the readings folder.')
    problems.add_argument('-w', '--window-size', nargs='+', required=False,
                          help='List of window sizes values to benchmark.')
    problems.add_argument('-r', '--resample-rule', nargs='+', required=False,
                          help='List of resample rule to benchmark.')
    problems.add_argument('-o', '--output', type=str,
                          help='Output path where to save the generated problems.')
    problems.add_argument('-s', '--signals', type=str,
                          help='Path to csv file that has signal_id column to use as the signal')

    return parser


def main():
    # Parse args
    parser = _get_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()
    args.action(args)


if __name__ == '__main__':
    warnings.filterwarnings("ignore")
    main()



# -*- coding: utf-8 -*-

import getpass
import json
import logging
from datetime import datetime

from pymongo import MongoClient

from greenguard.utils import remove_dots, restore_dots

LOGGER = logging.getLogger(__name__)


class MongoDB(object):

    def __init__(self, database=None, config=None, **kwargs):
        if config:
            with open(config, 'r') as f:
                config = json.load(f)
        else:
            config = kwargs

        host = config.get('host', 'localhost')
        port = config.get('port', 27017)
        user = config.get('user')
        password = config.get('password')
        database = database or config.get('database', 'test')
        auth_database = config.get('auth_database', 'admin')

        if user and not password:
            password = getpass.getpass(prompt='Please insert database password: ')

        client = MongoClient(
            host=host,
            port=port,
            username=user,
            password=password,
            authSource=auth_database
        )

        LOGGER.info("Setting up a MongoClient %s", client)

        self._db = client[database]

    def load_template(self, template_name):
        match = {
            'name': template_name
        }

        cursor = self._db.templates.find(match)
        templates = list(cursor.sort('insert_ts', -1).limit(1))

        if templates:
            return restore_dots(templates[0])

    def insert_template(self, template):
        if 'name' not in template:
            raise ValueError("Templates need to have a name key")

        template['insert_ts'] = datetime.utcnow()
        template = remove_dots(template)

        self._db.templates.insert_one(template)

    def insert_pipeline(self, candidate, score, dataset, table, column):

        pipeline = candidate.to_dict()

        pipeline['score'] = score
        pipeline['dataset'] = dataset
        pipeline['table'] = table
        pipeline['column'] = column
        pipeline['insert_ts'] = datetime.utcnow()

        pipeline = remove_dots(pipeline)

        self._db.pipelines.insert_one(pipeline)



import os
from random import random

import pandas as pd


def load_results(files):
    problems_results = dict()
    for filename in files:
        problem = os.path.basename(filename).replace('.csv', '')
        problems_results[problem] = pd.read_csv(filename).round(6)

    return problems_results


def get_wins_by_problems(results):
    df = results.groupby('problem_name')['template', 'window_size', 'resample_rule', 'fpr_threshold=0.5']
    df = df.apply(max)
    df = df.rename(columns={'fpr_threshold=0.5': 'score'})

    return df


def get_exclusive_wins(scores, column, pivot_columns=['window_size', 'resample_rule']):
    summary = {}
    for problem in scores.problem_name.unique():
        df = scores[scores['problem_name'] == problem]
        df['wr'] = df.apply(
            lambda row: '{}_{}_{}'.format(row[pivot_columns[0]], row[pivot_columns[1]], random()), axis=1)
        df = df.pivot(index='wr', columns=column, values='fpr_threshold=0.5')

        is_winner = df.T.rank(method='min', ascending=False) == 1
        num_winners = is_winner.sum()
        is_exclusive = num_winners == 1
        is_exclusive_winner = is_winner & is_exclusive
        summary[problem] = is_exclusive_winner.sum(axis=1)

    summary_df = pd.DataFrame(summary)
    summary_df.index.name = 'template'
    columns = summary_df.columns.sort_values(ascending=False)
    return summary_df[columns]


def add_sheet(dfs, name, writer, cell_fmt, index_fmt, header_fmt):
    startrow = 0
    widths = [0]
    if not isinstance(dfs, dict):
        dfs = {None: dfs}

    for df_name, df in dfs.items():
        df = df.reset_index()
        startrow += bool(df_name)
        df.to_excel(writer, sheet_name=name, startrow=startrow + 1, index=False, header=False)

        worksheet = writer.sheets[name]

        if df_name:
            worksheet.write(startrow - 1, 0, df_name, index_fmt)
            widths[0] = max(widths[0], len(df_name))

        for idx, column in enumerate(df.columns):
            worksheet.write(startrow, idx, column, header_fmt)
            width = max(len(column), *df[column].astype(str).str.len()) + 1
            if len(widths) > idx:
                widths[idx] = max(widths[idx], width)
            else:
                widths.append(width)

        startrow += len(df) + 2

    for idx, width in enumerate(widths):
        fmt = cell_fmt if idx else index_fmt
        worksheet.set_column(idx, idx, width + 1, fmt)


def write_results(results, output):
    writer = pd.ExcelWriter(output, engine='xlsxwriter')
    cell_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10"
    })
    index_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10",
        "bold": True,
    })
    header_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10",
        "bold": True,
        "bottom": 1
    })

    if isinstance(results, dict):
        results = pd.concat(list(results.values()), ignore_index=True)

    window = get_exclusive_wins(results, 'window_size', ['window_size', 'fpr_threshold=0.5'])

    resample_pivots = ['resample_rule', ['problem_name', 'fpr_threshold=0.5']]
    resample = get_exclusive_wins(results, 'resample_rule', resample_pivots)

    summary = {
        'Best pipeline by Problem': get_wins_by_problems(results),
        'Rankings - Number of wins': get_exclusive_wins(results, 'template'),
        'Resample Rule': resample,
        'Window Size': window
    }
    add_sheet(summary, 'Summary', writer, cell_fmt, index_fmt, header_fmt)

    for problem in results['problem_name'].unique():
        add_sheet(
            results[results['problem_name'] == problem],
            problem,
            writer,
            cell_fmt,
            index_fmt,
            header_fmt
        )

    writer.save()



# -*- coding: utf-8 -*-

"""Top-level package for GreenGuard."""

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '1.0.0'

import os

from greenguard.pipeline import GreenGuardPipeline, get_pipelines

_BASE_PATH = os.path.abspath(os.path.dirname(__file__))
MLBLOCKS_PIPELINES = os.path.join(_BASE_PATH, 'pipelines')
MLBLOCKS_PRIMITIVES = os.path.join(_BASE_PATH, 'primitives')


__all__ = (
    'GreenGuardPipeline',
    'get_pipelines',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `greenguard.pipeline` module."""
from unittest import TestCase
from unittest.mock import patch

import pandas as pd

from greenguard.pipeline import GreenGuardPipeline


class TestGreenGuardPipeline(TestCase):

    def _get_data(self):
        target_times = pd.DataFrame({
            'turbine_id': ['T001'],
            'cutoff_time': [pd.Timestamp('2010-01-01')],
            'target': [1]
        })
        readings = pd.DataFrame({
            'turbine_id': ['T001'],
            'timestamp': [pd.Timestamp('2010-01-01')],
            'signal_id': ['S1'],
            'value': [0.1]
        })
        return target_times, readings

    @patch('greenguard.pipeline.MLPipeline')
    @patch('greenguard.pipeline.load_pipeline')
    def test_fit(self, load_pipeline_mock, mlpipeline_mock):
        load_pipeline_mock.return_value = dict()

        # Run
        instance = GreenGuardPipeline('a_pipeline', 'accuracy')
        target_times, readings = self._get_data()
        instance.fit(target_times, readings)

        # Asserts
        assert instance.fitted

    @patch('greenguard.pipeline.MLPipeline')
    @patch('greenguard.pipeline.load_pipeline')
    def test_predict(self, load_pipeline_mock, mlpipeline_mock):
        load_pipeline_mock.return_value = dict()

        # Run
        instance = GreenGuardPipeline('a_pipeline', 'accuracy')
        instance.fitted = True
        target_times, readings = self._get_data()
        instance.predict(target_times, readings)



# -*- coding: utf-8 -*-
import logging

import numpy as np
from sklearn.metrics import (accuracy_score, f1_score, mean_absolute_error,
                             mean_squared_error, roc_curve, roc_auc_score, r2_score)

LOGGER = logging.getLogger(__name__)


def f1_macro(exp, obs):
    return f1_score(exp, obs, average='macro')


def threshold_score(ground_truth, probabilities, tpr):
    roc_fpr, roc_tpr, roc_threshold = roc_curve(ground_truth, probabilities, pos_label=1)
    try:
        index = np.where(roc_tpr >= tpr)[0][0]
    except:
        LOGGER.warn('Could not find a threshold that satisfies the requested True Positive Rate')
        index = -1

    return roc_threshold[index]


def tpr_score(ground_truth, probabilities, threshold):
    roc_fpr, roc_tpr, roc_threshold = roc_curve(ground_truth, probabilities, pos_label=1)
    try:
        index = np.where(roc_threshold >= threshold)[0][0]
    except:
        LOGGER.warn('Could not find a tpr that satisfies the requested threshold')
        index = -1

    return roc_tpr[index]


def fpr_score(ground_truth, probabilities, tpr=None, threshold=None):
    """Compute the False Positive Rate associated with the given True Positive Rate.

    This metric computes the False Positive Rate that needs to be assumed in order
    to achieve the desired True Positive Rate.
    The metric is computed by finding the minimum necessary threshold to ensure
    that the TPR is satisfied and then computing the associated FPR. The final output
    is 1 minus the found FPR to produce a maximization score between 0 and 1.

    Args:
        ground_truth (numpy.ndarray):
            ``numpy.ndarray`` of the known values for the given predictions.
        probabilities (numpy.ndarray):
            ``numpy.ndarray`` with the generated predictions in probability.
        tpr (float):
            ``float`` value representing the percentage of True Positive Rate
            to be satisfied.

    Returns:
        float:
            Value between 0 and 1, where bigger is better.
    """
    roc_fpr, roc_tpr, roc_threshold = roc_curve(ground_truth, probabilities, pos_label=1)
    try:
        if tpr:
            index = np.where(roc_tpr >= tpr)[0][0]
        elif threshold:
            index = np.where(roc_threshold >= threshold)[0][0]

    except:
        LOGGER.warn('Could not find a threshold that satisfies the requested True Positive Rate')
        index = -1

    return 1 - roc_fpr[index]


METRICS = {
    'accuracy': (accuracy_score, False),
    'f1': (f1_score, False),
    'f1_macro': (f1_macro, False),
    'r2': (r2_score, False),
    'mse': (mean_squared_error, True),
    'mae': (mean_absolute_error, True),
    'fpr': (fpr_score, False),
    'roc_auc_score': (roc_auc_score, False)
}



# -*- coding: utf-8 -*-

import logging
import os

import pandas as pd

LOGGER = logging.getLogger(__name__)

S3_URL = 'https://d3-ai-greenguard.s3.amazonaws.com/'
DEMO_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'demo')


def _load_or_download(filename, dates):
    filename += '.csv.gz'
    file_path = os.path.join(DEMO_PATH, filename)
    if os.path.exists(file_path):
        return pd.read_csv(file_path, compression='gzip', parse_dates=[dates])

    os.makedirs(DEMO_PATH, exist_ok=True)
    url = S3_URL + filename

    LOGGER.info('Downloading %s from %s', filename, url)
    data = pd.read_csv(url, compression='gzip', parse_dates=[dates])
    data.to_csv(file_path, index=False, compression='gzip')

    return data


def load_demo(load_readings=True):
    """Load the demo included in the GreenGuard project.

    The first time that this function is executed, the data will be downloaded
    and cached inside the `greenguard/demo` folder.
    Subsequent calls will load the cached data instead of downloading it again.

    Returns:
        tuple[pandas.DataFrame]:
            target_times and readings tables
    """
    target_times = _load_or_download('target_times', 'cutoff_time')
    if load_readings:
        readings = _load_or_download('readings', 'timestamp')
        return target_times, readings

    return target_times


def generate_raw_readings(output_path='demo'):
    """Generate raw readings based on the demo data.

    Args:
        path (str):
            Path where the readings will be generated.
    """
    target_times, readings = load_demo()

    for turbine_id in target_times.turbine_id.unique():
        turbine_path = os.path.join(output_path, turbine_id)
        os.makedirs(turbine_path, exist_ok=True)
        data = readings[readings.turbine_id == turbine_id]
        for month in range(1, 13):
            month_data = data[data.timestamp.dt.month == month].copy()
            month_data['timestamp'] = month_data['timestamp'].dt.strftime('%m/%d/%y %H:%M:%S')
            month_path = os.path.join(turbine_path, '2013-{:02d}.csv'.format(month))
            LOGGER.info('Generating file %s', month_path)
            month_data[['signal_id', 'timestamp', 'value']].to_csv(month_path, index=False)

    return target_times



import numpy as np

from greenguard.metrics import fpr_score


def test_fpr_score_perfect_scenario():
    truth = [0, 0, 0, 1, 1, 1]
    false_probs = [0.2, 0.4, 0.6]
    true_probs = [0.8, 0.7, 0.9]
    probs = np.concatenate([false_probs, true_probs])
    score = fpr_score(truth, probs, tpr=1)
    assert score == 1


def test_fpr_score_predict_over_half():
    truth = [0, 0, 0, 0, 1, 1, 1, 1]
    false_probs = [0.1, 0.2, 0.4, 0.6]
    true_probs = [0.5, 0.7, 0.8, 0.9]
    probs = np.concatenate([false_probs, true_probs])
    score = fpr_score(truth, probs, tpr=1)
    assert score == 0.75


def test_fpr_score_predict_half():
    truth = [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1]
    false_probs = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
    true_probs = [0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    probs = np.concatenate([false_probs, true_probs])
    score = fpr_score(truth, probs, tpr=1)
    assert score == 0.5


def test_fpr_score_predict_one_third():
    truth = [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1]
    false_probs = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
    true_probs = [0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    probs = np.concatenate([false_probs, true_probs])
    score = fpr_score(truth, probs, tpr=1)
    assert round(score, 4) == 0.3333



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

try:
    with open('README.md', encoding='utf-8') as readme_file:
        readme = readme_file.read()
except IOError:
    readme = ''

try:
    with open('HISTORY.md', encoding='utf-8') as history_file:
        history = history_file.read()
except IOError:
    history = ''

install_requires = [
    'baytune>=0.4.0,<0.5',
    'mlprimitives>=0.3.0,<0.4',
    'mlblocks>=0.4.0,<0.5',
    'pymongo>=3.7.2,<4',
    'scikit-learn>=0.21',
    'tqdm<4.50.0,>=4.36.1',
    'cloudpickle>=1.6,<2',
    'scipy>=1.0.1,<2',
    'numpy<1.19.0,>=1.16.0',
    'pandas>=1,<2',
    'partd>=1.1.0,<2',
    'fsspec>=0.8.5,<0.9',
    'dask>=2.6.0,<3',
    'distributed>=2.6.0,<3',
    'h5py<2.11.0,>=2.10.0',  # fix tensorflow requirement
    'Keras>=2.4',
    'tabulate>=0.8.3,<0.9',
    'xlsxwriter>=1.3.6<1.4',
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
    'jupyter>=1.0.0,<2',
    'rundoc>=0.4.3,<0.5',
]

development_requires = [
    # general
    'bumpversion>=0.5.3,<0.6',
    'pip>=9.0.1',
    'watchdog>=0.8.3,<0.11',

    # docs
    'm2r>=0.2.0,<0.3',
    'nbsphinx>=0.5.0,<0.7',
    'Sphinx>=1.7.1,<3',
    'sphinx_rtd_theme>=0.2.4,<0.5',
    'autodocsumm>=0.1.10',

    # style check
    'flake8>=3.7.7,<4',
    'isort>=4.3.4,<5',

    # fix style issues
    'autoflake>=1.1,<2',
    'autopep8>=1.4.3,<2',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',
    'importlib-metadata<2,>=0.12',
]

setup(
    author='MIT Data To AI Lab',
    author_email='dailabmit@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description='AutoML for Renewable Energy Industries.',
    entry_points={
        'mlblocks': [
            'pipelines=greenguard:MLBLOCKS_PIPELINES',
            'primitives=greenguard:MLBLOCKS_PRIMITIVES'
        ],
    },
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='wind machine learning greenguard',
    license='MIT license',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    name='greenguard',
    packages=find_packages(include=['greenguard', 'greenguard.*']),
    python_requires='>=3.6,<3.9',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/D3-AI/GreenGuard',
    version='1.0.0',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

import gc
import json
import logging
import os
import pickle
import tempfile
from copy import deepcopy
from hashlib import md5

import cloudpickle
import keras
import numpy as np
from btb import BTBSession
from btb.tuning import Tunable
from mlblocks import MLPipeline
from mlblocks.discovery import load_pipeline
from mlprimitives.adapters.keras import Sequential
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import KFold, StratifiedKFold

from greenguard.metrics import METRICS

LOGGER = logging.getLogger(__name__)


PIPELINES_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'pipelines'))


# Patch Keras to save on disk without a model trained
def __getstate__(self):
    state = self.__dict__.copy()
    if 'model' in state:
        with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=True) as fd:
            keras.models.save_model(state.pop('model'), fd.name, overwrite=True)
            state['model_str'] = fd.read()

    return state


def __setstate__(self, state):
    if 'model_str' in state:
        with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=True) as fd:
            fd.write(state.pop('model_str'))
            fd.flush()

            state['model'] = keras.models.load_model(fd.name)

    self.__dict__ = state


Sequential.__getstate__ = __getstate__
Sequential.__setstate__ = __setstate__


def get_pipelines(pattern='', path=False, pipeline_type='classes'):
    """Get the list of available pipelines.

    Optionally filter the names using a patter or obtain
    the paths to the pipelines alongside their name.

    Args:
        pattern (str):
            Pattern to search for in the pipeline names
        path (bool):
            Whether to return a dictionary containing the pipeline
            paths instead of only a list with the names.
        pipeline_type (str):
            The pipeline category to filter by (`classes`, `probability` and `unstacked`).
            Defaults to `classes`.

    Return:
        list or dict:
            List of available and matching pipeline names.
            If `path=True`, return a dict containing the pipeline
            names as keys and their absolute paths as values.
    """
    pipelines = dict()
    pipelines_dir = os.path.join(PIPELINES_DIR, pipeline_type)

    for filename in os.listdir(pipelines_dir):
        if filename.endswith('.json') and pattern in filename:
            name = os.path.basename(filename)[:-len('.json')]
            name = f'{pipeline_type}.{name}'
            pipeline_path = os.path.join(pipelines_dir, filename)
            pipelines[name] = pipeline_path

    if not path:
        pipelines = list(pipelines)

    return pipelines


def generate_init_params(template_names, init_params):
    """Generate init_params dicts.

    Args:
        template_names (list):
            List of templates.
        init_params (list or dict):
            Initialization parameters for the templates.

    Returns:
        Dict that contains one entry for each template with a dict indicating
        the init_params to use with that template.
    """
    if not init_params:
        init_params = {}
    elif isinstance(init_params, list):
        init_params = dict(zip(template_names, init_params))

    if not any(name in init_params for name in template_names):
        return {
            name: deepcopy(init_params)
            for name in template_names
        }
    else:
        return {
            name: deepcopy(init_params.get(name, {}))
            for name in template_names
        }


def generate_preprocessing(templates_names, preprocessing):
    """Generate preprocessing dict.

    Args:
        template_names (list):
            List of templates.
        preprocessing (int, list or dict):
            Number of preprocessing steps to be used.

    Returns:
        preprocessing (dict):
            Contains one entry for each template and an integer indicating the
            number of preprocessing steps for that template.
    """
    if isinstance(preprocessing, int):
        preprocessing = {name: preprocessing for name in templates_names}
    else:
        if isinstance(preprocessing, list):
            preprocessing = dict(zip(templates_names, preprocessing))

        preprocessing = {
            name: preprocessing.get(name, 0)
            for name in templates_names
        }
    return preprocessing


SELF_THRESHOLD = object()


class GreenGuardPipeline(object):
    """Main Machine Learning component in the GreenGuard project.

    The ``GreenGuardPipeline`` represents the abstraction of a Machine
    Learning pipeline architecture specialized on the GreenGuard data
    format.

    In order to use it, an MLBlocks pipeline template needs to be given,
    alongside information about how to evaluate its performance using
    cross validation.

    Attributes:
        template (MLPipeline):
            MLPipeline instance used as the template for tuning.
        template_name:
            Name of the template being used.
        fitted (bool):
            Whether this GreenGuardPipeline has already been fitted or not.
        steps (list):
            List of primitives that compose this template.
        preprocessing (list):
            List of preprocessing steps. These steps have no learning stage
            and are executed only once on the complete training dataset, before
            partitioning it for cross validation.
        static (list):
            List of static steps. These are all the steps in the pipeline that
            come after the preprocessing ones but have no hyperparameters.
            These are executed on each cross validation split only once, when
            the data is partitioned, and their output is cached to be reused
            later on at every tuning iteration.
        tunable (list):
            List of steps that have hyperparameters and will be tuned during
            the tuning loop.

    Args:
        templates (str, MLPipeline or list):
            Template to use. If a ``str`` is given, load the corresponding
            ``MLPipeline``. Also can be a list combining both.
        metric (str or function):
            Metric to use. If an ``str`` is given it must be one of the metrics
            defined in the ``greenguard.metrics.METRICS`` dictionary.
        cost (bool):
            Whether the metric is a cost function (the lower the better) or not.
            Defaults to ``False``.
        init_params (dict or list):
            There are three possible values for init_params:

                * Init params ``dict``: It will be used for all templates.
                * ``dict`` with the name of the template as a key and dictionary with its
                  init params.
                * ``list``: each value will be assigned to the corresponding position of
                  self.templates.

            Defaults to ``None``.
        stratify (bool):
            Whether to stratify the data when partitioning for cross validation.
            Defaults to ``True``.
        cv_splits (int):
            Number of cross validation folds to use. Defaults to ``5``.
        shuffle (bool):
            Whether to shuffle the data when partitioning for cross validation.
            Defaults to ``True``.
        random_state (int or RandomState):
            random state to use for the cross validation partitioning.
            Defaults to ``0``.
        preprocessing (int, dict or list):
            There are three possible values for preprocessing:

                * ``int``: the value will be used for all templates.
                * ``dict`` with the template name as a key and a number as a value, will
                  be used for that template.
                * ``list``: each value will be assigned to the corresponding position of
                  self.templates.

            Defaults to ``0``.
        cache_path (str):
            If given, cache the generated cross validation splits in this folder.
            Defatuls to ``None``.
        threshold (float):
            If ``None``, return the raw predictions as given by the pipeline. If not ``None``,
            use the given value as a threshold to convert the predicted probabilities into
            a binary output that indicates whether the probability is above the threshold (not
            strict) or below the threshold (strict). Defaults to ``None``.
    """

    template = None
    template_name = None
    fitted = False
    cv_score = None

    _cv_class = None
    _metric = None
    _cost = False
    _tuner = None
    _pipeline = None
    _static = None
    _init_params = None
    _preprocessing = None

    def _get_cv(self, stratify, cv_splits, shuffle, random_state):
        if stratify:
            cv_class = StratifiedKFold
        else:
            cv_class = KFold

        return cv_class(n_splits=cv_splits, shuffle=shuffle, random_state=random_state)

    def _set_hyperparameters(self, new_hyperparameters):
        self._hyperparameters = deepcopy(new_hyperparameters)

    def _set_template(self, template_name):
        self.template_name = template_name
        self.template = self._template_dicts[self.template_name]

    @staticmethod
    def _update_params(old, new):
        for name, params in new.items():
            if '#' not in name:
                name = name + '#1'

            block_params = old.setdefault(name, dict())
            for param, value in params.items():
                block_params[param] = value

    def _count_static_steps(self, pipeline):
        tunable_hyperparams = pipeline.get_tunable_hyperparameters()
        for index, block_name in enumerate(pipeline.blocks.keys()):
            if tunable_hyperparams[block_name]:
                return index

        return 0

    def _get_templates(self, templates):
        template_dicts = dict()
        template_names = list()
        for template in templates:
            if isinstance(template, str):
                template_name = template
                if os.path.isfile(template):
                    with open(template, 'r') as json_file:
                        template = json.load(json_file)
                else:
                    template = deepcopy(load_pipeline(template_name))
            else:
                template_name = md5(json.dumps(template)).digest()

            template_dicts[template_name] = template
            template_names.append(template_name)

        return template_names, template_dicts

    def _build_pipeline(self):
        self._pipeline = MLPipeline(self.template)

        if self._hyperparameters:
            self._pipeline.set_hyperparameters(self._hyperparameters)

        self.fitted = False

    def __init__(self, templates, metric='accuracy', cost=False, init_params=None,
                 stratify=True, cv_splits=5, shuffle=True, random_state=0, preprocessing=0,
                 cache_path=None, threshold=None):

        if isinstance(metric, str):
            metric, cost = METRICS[metric]

        self._metric = metric
        self._cost = cost
        self._cv = self._get_cv(stratify, cv_splits, shuffle, random_state)
        self.cv_score = np.inf if cost else -np.inf
        self.threshold = threshold

        if not isinstance(templates, list):
            templates = [templates]

        self.templates = templates
        self._template_names, self._template_dicts = self._get_templates(templates)
        self._default_init_params = {}
        self._init_params = generate_init_params(self._template_names, init_params)

        for name, template in self._template_dicts.items():
            init_params = self._init_params.get(name, self._default_init_params)
            template_params = template.setdefault('init_params', {})
            self._update_params(template_params, init_params)

        self._preprocessing = generate_preprocessing(self._template_names, preprocessing)
        self._set_template(self._template_names[0])
        self._hyperparameters = dict()
        self._build_pipeline()
        self._cache_path = cache_path
        if cache_path:
            os.makedirs(cache_path, exist_ok=True)

    def get_hyperparameters(self):
        """Get the current hyperparameters.

        Returns:
            dict:
                Current hyperparameters.
        """
        return deepcopy(self._hyperparameters)

    def _is_better(self, score):
        if self._cost:
            return score < self.cv_score

        return score > self.cv_score

    def _generate_splits(self, template_name, target_times, readings, turbines=None):
        template = self._template_dicts.get(template_name)
        pipeline = MLPipeline(template)
        preprocessing = self._preprocessing.get(template_name)
        static = self._count_static_steps(pipeline)
        X = target_times[['turbine_id', 'cutoff_time']]
        y = target_times['target']

        if preprocessing:
            if preprocessing > static:
                raise ValueError('Preprocessing cannot be bigger than static')

            LOGGER.debug('Running %s preprocessing steps', preprocessing)
            context = pipeline.fit(X=X, y=y, readings=readings,
                                   turbines=turbines, output_=preprocessing - 1)
            del context['X']
            del context['y']
            gc.collect()

        else:
            context = {
                'readings': readings,
                'turbines': turbines,
            }

        splits = list()
        for fold, (train_index, test_index) in enumerate(self._cv.split(X, y)):
            LOGGER.debug('Running static steps for fold %s', fold)
            gc.collect()
            X_train, X_test = X.iloc[train_index], X.iloc[test_index]
            y_train, y_test = y.iloc[train_index], y.iloc[test_index]

            pipeline = MLPipeline(template)
            fit = pipeline.fit(X_train, y_train, output_=static - 1,
                               start_=preprocessing, **context)
            predict = pipeline.predict(X_test, output_=static - 1,
                                       start_=preprocessing, **context)

            split = (fold, pipeline, fit, predict, y_test, static)

            if self._cache_path:
                split_name = '{}_{}.pkl'.format(template_name, fold)
                split_path = os.path.join(self._cache_path, split_name)
                os.makedirs(os.path.dirname(split_path), exist_ok=True)

                with open(split_path, 'wb') as split_file:
                    pickle.dump(split, split_file)

                split = split_path

            splits.append(split)

        gc.collect()
        return splits

    def _cross_validate(self, template_splits, hyperparams):
        scores = []
        for split in template_splits:
            gc.collect()
            if self._cache_path:
                with open(split, 'rb') as split_file:
                    split = pickle.load(split_file)

            fold, pipeline, fit, predict, y_test, static = split

            LOGGER.debug('Scoring fold %s', fold)
            pipeline.set_hyperparameters(hyperparams)
            pipeline.fit(start_=static, **fit)
            predictions = pipeline.predict(start_=static, **predict)

            score = self._metric(y_test, predictions)
            LOGGER.debug('Fold fold %s score: %s', fold, score)
            scores.append(score)

        return np.mean(scores)

    def _make_btb_scorer(self, target_times, readings, turbines):
        splits = {}

        def scorer(template_name, config):
            template_splits = splits.get(template_name)
            if template_splits is None:
                template_splits = self._generate_splits(
                    template_name, target_times, readings, turbines)

                splits[template_name] = template_splits

            cv_score = self._cross_validate(template_splits, config)
            if self._is_better(cv_score):
                _config = '\n'.join('      {}: {}'.format(n, v) for n, v in config.items())
                LOGGER.info(('New configuration found:\n'
                             '  Template: %s \n'
                             '    Hyperparameters: \n'
                             '%s'), template_name, _config)

                self.cv_score = cv_score
                self._set_template(template_name)
                self._set_hyperparameters(config)
                self._build_pipeline()

            return cv_score

        return scorer

    def cross_validate(self, target_times, readings, turbines,
                       template_name=None, hyperparams=None):
        """Compute cross validation score using the given data.

        If the splits have not been previously computed, compute them now.
        During this computation, the data is partitioned using the indicated
        cross validation parameters and later on processed using the
        pipeline static steps.

        The results of the fit and produce executions are cached and reused
        in subsequent calls to this method.

        Args:
            X (pandas.DataFrame):
                ``target_times`` data, without the ``target`` column.
                Only needed if the splits have not been previously computed.
            y (pandas.Series or numpy.ndarray):
                ``target`` vector corresponding to the passed ``target_times``.
                Only needed if the splits have not been previously computed.
            readings (pandas.DataFrame):
                ``readings`` table. Only needed if the splits have not been
                previously computed.
            turbines (pandas.DataFrame):
                ``turbines`` table. Only needed if the splits have not been
                previously computed.
            params (dict):
                hyperparameter values to use.

        Returns:
            float:
                Computed cross validation score. This score is the average
                of the scores obtained accross all the cross validation folds.
        """
        if not template_name:
            template_name = self.template_name
            if hyperparams is None:
                hyperparams = self.get_hyperparameters()

        elif hyperparams is None:
            hyperparams = {}

        template_splits = self._generate_splits(template_name, target_times, readings, turbines)
        return self._cross_validate(template_splits, hyperparams)

    @classmethod
    def _get_tunables(cls, template_dicts):
        tunables = {}
        for name, template in template_dicts.items():
            pipeline = MLPipeline(template)
            pipeline_tunables = pipeline.get_tunable_hyperparameters(flat=True)
            tunables[name] = Tunable.from_dict(pipeline_tunables)

        return tunables

    def tune(self, target_times, readings, turbines=None):
        """Create a tuning session object that tunes and selects the templates.

        Args:
            target_times (pandas.DataFrame):
                ``target_times`` table, containing the ``turbine_id``, ``cutoff_time``
                and ``target`` columns.
                Only needed if the splits have not been previously computed.
            readings (pandas.DataFrame):
                ``readings`` table. Only needed if the splits have not been
                previously computed.
            turbines (pandas.DataFrame):
                ``turbines`` table. Only needed if the splits have not been
                previously computed.
        """
        scoring_function = self._make_btb_scorer(target_times, readings, turbines)
        tunables = self._get_tunables(self._template_dicts)
        return BTBSession(tunables, scoring_function, maximize=not self._cost)

    def fit(self, target_times=None, readings=None, turbines=None,
            start_=None, output_=None, **kwargs):
        """Fit this pipeline to the given data.

        Args:
            target_times (pandas.DataFrame):
                ``target_times`` table, containing the ``turbine_id``, ``cutoff_time``
                and ``target`` columns.
            readings (pandas.DataFrame):
                ``readings`` table.
            turbines (pandas.DataFrame):
                ``turbines`` table.
        """
        if target_times is None:
            X = kwargs.pop('X')
            y = kwargs.pop('y')
        else:
            X = target_times[['turbine_id', 'cutoff_time']]
            y = target_times['target']

        out = self._pipeline.fit(X, y, readings=readings, turbines=turbines,
                                 start_=start_, output_=output_, **kwargs)

        if output_ is None:
            self.fitted = True

        return out

    def predict(self, target_times=None, readings=None, turbines=None,
                start_=None, output_='default', threshold=SELF_THRESHOLD, **kwargs):
        """Make predictions using this pipeline.

        Args:
            target_times (pandas.DataFrame):
                ``target_times`` table, containing the ``turbine_id``, ``cutoff_time``
                and ``target`` columns.
            readings (pandas.DataFrame):
                ``readings`` table.
            turbines (pandas.DataFrame):
                ``turbines`` table.
            threshold (float):
                If not given, use the threshold specified upon instance creation in the
                ``__init__``. If ``None``, return the raw predictions as given by the pipeline.
                If not ``None``, use the given value as a threshold to convert the predicted
                probabilities into a binary output that indicates whether the probability is above
                the threshold (not strict) or below the threshold (strict).
                Defaults to ``self.threshold``.

        Returns:
            numpy.ndarray:
                Vector of predictions.
        """
        if not self.fitted:
            raise NotFittedError()

        X = target_times[['turbine_id', 'cutoff_time']]
        predictions = self._pipeline.predict(X, readings=readings, turbines=turbines,
                                             start_=start_, output_=output_, **kwargs)
        if threshold is SELF_THRESHOLD:
            threshold = self.threshold

        if threshold is not None:
            predictions = predictions >= threshold

        return predictions

    def save(self, path):
        """Serialize and save this pipeline using cloudpickle.

        Args:
            path (str):
                Path to the file where the pipeline will be saved.
        """
        with open(path, 'wb') as pickle_file:
            cloudpickle.dump(self, pickle_file)

    @classmethod
    def load(cls, path):
        """Load a previously saved pipeline from a file.

        Args:
            path (str):
                Path to the file where the pipeline is saved.

        Returns:
            GreenGuardPipeline:
                Loaded GreenGuardPipeline instance.
        """
        with open(path, 'rb') as pickle_file:
            return cloudpickle.load(pickle_file)
